/* Copyright (c) 2014 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @defgroup ble_sdk_app_proximity_main main.c
 * @{
 * @ingroup ble_sdk_app_proximity_eval
 * @brief Proximity Application main file.
 *
 * This file contains is the source code for a sample proximity application using the
 * Immediate Alert, Link Loss and Tx Power services.
 *
 * This application would accept pairing requests from any peer device.
 *
 * It demonstrates the use of fast and slow advertising intervals.
 */

#include <stdint.h>
#include <string.h>

#include "nordic_common.h"
#include "nrf.h"
#include "nrf_delay.h"
#include "nrf_soc.h"
#include "app_error.h"
#include "softdevice_handler.h"
#include "app_timer.h"
#include "app_util.h"
#include "bsp.h"

#include "fds.h"
#include "fstorage.h"

#include "sdk_config.h"

#define NRF_LOG_MODULE_NAME "MAIN"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"

#define APP_TIMER_PRESCALER                 0                                            /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE             6                                            /**< Size of timer operation queues. */

#define DEAD_BEEF                           0xDEADBEEF                                   /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */


#define FILE_ID     0x1111
#define REC_KEY     0x2222


uint16_t m_file_id = FILE_ID;
uint16_t m_rec_key = REC_KEY;

static volatile uint8_t write_flag = 0;
static volatile uint8_t init_flag = 0;
/*
 * Local reference to last record id accessed
 */
#define NVM_INVALID_EVENT      0xFF
#define NVM_INVALID_RECORD     0xFFFFFFFF

static uint32_t     volatile m_nvm_mgt_last_record_id = NVM_INVALID_RECORD;
static fds_evt_id_t volatile m_nvm_mgt_last_event     = NVM_INVALID_EVENT;



/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num   Line number of the failing ASSERT call.
 * @param[in] file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module. This creates and starts application timers.
 */
static void timers_init(void)
{
    // Initialize timer module.
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
}

/**@brief Function for dispatching a system event to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in] sys_evt  System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt)
{
    fs_sys_event_handler(sys_evt);
}



/**@brief Function for handling events from the BSP module.
 *
 * @param[in]   event   Event generated when button is pressed.
 */
static void bsp_event_handler(bsp_event_t event)
{
    //    uint32_t err_code;

    switch (event)
    {

        case BSP_EVENT_KEY_0:
            break;//BSP_EVENT_KEY_0

        case BSP_EVENT_KEY_1: // STOP_ALERTING_BUTTON_ID
            break;//BSP_EVENT_KEY_1

        default:
            break;
    }
}


/**@brief Function for initializing buttons and leds.
 *
 */
static void buttons_leds_init()
{
    uint32_t err_code = bsp_init(BSP_INIT_LED | BSP_INIT_BUTTONS,
            APP_TIMER_TICKS(100, APP_TIMER_PRESCALER),
            bsp_event_handler);

    APP_ERROR_CHECK(err_code);

}


/**@brief Function for the Power manager.
*/
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();

    APP_ERROR_CHECK(err_code);
}


static void my_fds_evt_handler(fds_evt_t const * const p_fds_evt)
{
    // Store the current event as last event that happened
    m_nvm_mgt_last_event = p_fds_evt->id;
    NRF_LOG_DEBUG("FDS event : %d, result : 0x%04x\r\n", m_nvm_mgt_last_event, p_fds_evt->result);
    switch (p_fds_evt->id)
    {
        case FDS_EVT_INIT:
            if (p_fds_evt->result != FDS_SUCCESS)
            {
                // Initialization failed.
                NRF_LOG_ERROR("FDS init failed !\r\n");
            }
            else
            {
                // Managing an init flag so that nothing is done before FDS is ready...
                NRF_LOG_DEBUG("FDS init done !\r\n");
                m_nvm_mgt_last_record_id = NVM_INVALID_RECORD;
            }
            break;
        case FDS_EVT_WRITE:
            if (p_fds_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_DEBUG("FDS Write event, record key = 0x%04x, file id = 0x%04x\r\n",
                        p_fds_evt->write.record_key, p_fds_evt->write.file_id);
                m_nvm_mgt_last_record_id = p_fds_evt->write.record_id;
            }
            break;
        case FDS_EVT_UPDATE:
            if (p_fds_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_DEBUG("FDS Update event, record key = 0x%04x, file id = 0x%04x\r\n",
                        p_fds_evt->write.record_key, p_fds_evt->write.file_id);
                m_nvm_mgt_last_record_id = p_fds_evt->write.record_id;
            }
            break;
        case FDS_EVT_DEL_RECORD:
            if (p_fds_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_DEBUG("FDS delete record event\r\n");
                m_nvm_mgt_last_record_id = p_fds_evt->del.record_id;
            }
            break;
        case FDS_EVT_DEL_FILE:
            // TODO: Check whther we need to wait at some point for delete events
            // in any modules
            if (p_fds_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_DEBUG("FDS delete file event\r\n");
            }
            break;
        case FDS_EVT_GC:
            if (p_fds_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_DEBUG("FDS garbage collection event\r\n");
            }
            break;
        default:
            break;
    }

}
/**@brief       Function that waits for specific flag
 *
 *              TODO: Use power management function that we will implement
 *              for now calling directly sd_app_event_wait(); or __WFE()
 * @param       flag to look at
 */
void fds_test_wait_for(uint32_t record_id, fds_evt_id_t event)
{
    NRF_LOG_DEBUG("Waiting for event :  waited id is 0x%08x, current id is 0x%08x, waited event is %d, current event is %d\r\n",
                record_id, m_nvm_mgt_last_record_id, event, m_nvm_mgt_last_event);


    while((m_nvm_mgt_last_record_id != record_id) || (event != m_nvm_mgt_last_event))
    {
       __WFE();
    }
    // Reset values for next event to wait
    m_nvm_mgt_last_record_id = NVM_INVALID_RECORD;
    m_nvm_mgt_last_event = NVM_INVALID_EVENT;
    NRF_LOG_DEBUG("Out of wait\r\n");
}


static ret_code_t fds_test_init (void)
{

    ret_code_t ret = fds_register(my_fds_evt_handler);
    if (ret != FDS_SUCCESS)
    {
        return ret;

    }
    ret = fds_init();
    if (ret != FDS_SUCCESS)
    {
        return ret;
    }
    fds_test_wait_for(NVM_INVALID_RECORD, FDS_EVT_INIT);

    return NRF_SUCCESS;

}

/**
 * @brief       Function to create and store a file in NVM
 *
 * @param       file_id     File identifier
 * @param       record_id   Record key
 * @param       p_src_data  Pointer to memroy in RAM to copy from
 *                          WARNING: this pointer must be an aligned 4 address
 * @param       data_size   Size IN WORDS of the data to store
 */
ret_code_t fds_store_my_data(uint16_t file_id, uint16_t record_key,
        uint32_t * p_src_data, uint16_t data_word_size, uint32_t * record_id)
{
    fds_record_t        record;
    fds_record_chunk_t  record_chunk;
    fds_record_desc_t   record_desc;

    NRF_LOG_DEBUG("Store data from address 0x%08x, len = %d in words\r\n", (uint32_t) p_src_data, (uint32_t)data_word_size);
    NRF_LOG_DEBUG("File id is 0x%04x, record_key is 0x%04x\r\n", (uint32_t)file_id, (uint32_t)record_key);
    // Set up data.
    record_chunk.p_data         = p_src_data;
    record_chunk.length_words   = data_word_size;

    // Set up record.
    record.file_id              = file_id;
    record.key                  = record_key;
    record.data.p_chunks        = &record_chunk;
    record.data.num_chunks      = 1;

    ret_code_t ret = fds_record_write(&record_desc, &record);
    if (ret != FDS_SUCCESS)
    {
        return ret;
    }
    NRF_LOG_DEBUG("Writing Record ID = %d \r\n", record_desc.record_id);
    *record_id = record_desc.record_id;
    return NRF_SUCCESS;

}

/**@brief       Update an existing record with new data
 *
 * @details     This function needs a record ID to store new data
 *              It can be obtains from file id and record key using fds_get_data_record
 *              method below
 *
 * @param       fds_record_desc_t, pointer on the handle to the record to update
 * @param       p_src_data: Pointer uint32_t to the data to save
 * @param       data_word_size: Length of the data to store in WORDS
 */
ret_code_t fds_update_my_data(fds_record_desc_t * record_desc, uint16_t file_id, uint16_t rec_key,
        uint32_t* p_src_data, uint16_t data_word_size)
{
    ret_code_t err_code;
    fds_record_t        record;
    fds_record_chunk_t  record_chunk;

    // Set up data.
    record_chunk.p_data         = p_src_data;
    record_chunk.length_words   = data_word_size;
    record.data.p_chunks        = &record_chunk;
    record.data.num_chunks      = 1;

    // Set up record.
    record.file_id              = file_id;
    record.key                  = rec_key;

    // Update record
    err_code = fds_record_update(record_desc, &record);
    if (err_code != FDS_SUCCESS)
    {
        NRF_LOG_WARNING("Update data for record id 0x%04x has failed with error code 0x%02x\r\n",
                                      record_desc->record_id, err_code);
        return err_code;
    }
    NRF_LOG_DEBUG("Update Record ID = %d \r\n", record_desc->record_id);
    return NRF_SUCCESS;
}

/*
 * Test to validate basic functions for fds_storage layer, just above Nordic's FDS
 * (OK)
 */
static    uint32_t p_data[6] = {0x11111111, 0x22222222, 0x33333333, 0x44444444, 0x55555555, 0x66666666};
static    uint32_t p_udata[6]= {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff};
void fds_storage_update_test()
{
    uint16_t nb_update = 0;
    uint16_t w_size = 6;
    fds_record_desc_t record_desc;
    uint32_t record_id = 0;
    fds_find_token_t    ftok ={0};//Important, make sure you zero init the ftok token

    uint32_t val = 0x88889999;
    bool gc_called = false;

    NRF_LOG_DEBUG("RAM @ of p_data = 0x%08x\r\n", (uint32_t)p_data);
    NRF_LOG_DEBUG("RAM @ of p_udata= 0x%08x\r\n", (uint32_t)p_udata);

    // Create a file
    APP_ERROR_CHECK(fds_store_my_data(0x5555, 0x1212, (uint32_t *)p_data, w_size, &record_id));
    fds_test_wait_for(record_id, FDS_EVT_WRITE);

    NRF_LOG_DEBUG("Stored file 5555/1212: record_id = %d\r\n", record_id);

    APP_ERROR_CHECK(fds_record_find(0x5555, 0x1212, &record_desc, &ftok));

    while(nb_update < 60)
    {
        NRF_LOG_DEBUG("Update number %d\r\n", nb_update);
        // Change p_data
        p_data[2] = val++;
        p_data[5] = val++;
        // Update the record
        if(fds_update_my_data(&record_desc, 0x5555, 0x1212, p_data, w_size) == FDS_ERR_NO_SPACE_IN_FLASH)
        {
            if(!gc_called)
            {
                nrf_delay_ms(1000); // <- Those delays are there to allow debug log output
                NRF_LOG_DEBUG("MEMORY FULL!!! Calling garbage collector\r\n");
                fds_gc();
                fds_test_wait_for(NVM_INVALID_RECORD, FDS_EVT_GC);
                NRF_LOG_DEBUG("GC has been run, look at the memory!\r\n");
                nrf_delay_ms(15000);
                gc_called = true;
            }
            else
            {
                nrf_delay_ms(1000);// <- Those delays are there to allow debug log output
                NRF_LOG_DEBUG("\r\n\r\nGarbage collector already called once\r\n\r\n");
                nrf_delay_ms(1000);// <- Those delays are there to allow debug log output
                break;
            }
        }
        else
        {
            fds_test_wait_for(record_desc.record_id, FDS_EVT_UPDATE);
        }
        nb_update++;
    }
    APP_ERROR_CHECK(fds_record_find(0x5555, 0x1212, &record_desc, &ftok));
    // Calling GC to see whther it can acts without memory being full 
    fds_gc();
    fds_test_wait_for(NVM_INVALID_RECORD, FDS_EVT_GC);
    NRF_LOG_DEBUG("GC has been run, look at the memory!\r\n");
    APP_ERROR_CHECK(fds_record_find(0x5555, 0x1212, &record_desc, &ftok));
}

int main(void)
{
    uint32_t err_code;

    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;

    // Initialize.
    APP_ERROR_CHECK(NRF_LOG_INIT(NULL));
    NRF_LOG_INFO("\r\n\r\n\r\nInit logging done\r\n");

    // Initialize the SoftDevice handler module.
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    timers_init();
    buttons_leds_init();

    NRF_LOG_INFO("Start FDS test...\r\n");		

    fds_test_init();
    fds_storage_update_test();


    // Enter main loop.
    nrf_delay_ms(3000);// <- Those delays are there to allow debug log output
    NRF_LOG_DEBUG("\r\n");
    NRF_LOG_DEBUG("End of test\r\n");
    for (;;)
    {
        power_manage();
    }
}

/**
 * @}
 */
